from mmcc_framework import Response


def start(data, kb, context):
    return Response(kb, context, True, payload={"show_name": False, "show_age": False, "show_choose_name": True,
                                                "show_choose_nickname": True, "show_field": False,
                                                "field_contents": ""})


def insert_name(data, kb, context):
    if "name" in data and data["name"] != "":
        context["name"] = data["name"]
        return Response(kb, context, True, payload={"show_name": False, "show_age": True})
    return Response(kb, context, False, utterance=kb["insert_name_err"])


def insert_nickname(data, kb, context):
    if "nickname" in data and data["nickname"] != "":
        context["nickname"] = data["nickname"]
        return Response(kb, context, True, payload={"show_name": False, "show_age": True})
    return Response(kb, context, False, utterance=kb["insert_nickname_err"])


def insert_age(data, kb, context):
    if "age" in data:
        try:
            age = int(data["age"])
            if age < kb["insert_age_old"]:
                context["age"] = data["age"]
                if "name" in context:
                    name = "Name, " + context["name"]
                else:
                    name = "Nickname, " + context["nickname"]
                return Response(kb, context, True, utterance=f'{name} {context["age"]}',
                                payload={"show_age": False, "show_field": True,
                                         "field_contents": name + " " + context["age"]})
        except ValueError:
            pass
    return Response(kb, context, False, utterance=kb["insert_age_err"])


def name_nickname(data, kb, context):
    if "name_nickname" in data:
        if data["name_nickname"] == "name":
            return Response(kb, context, True, choice="insert_name",
                            payload={"show_choose_name": False, "show_choose_nickname": False, "show_name": True})
        if data["name_nickname"] == "nickname":
            return Response(kb, context, True, choice="insert_nickname",
                            payload={"show_choose_name": False, "show_choose_nickname": False})
    return Response(kb, context, False, utterance=kb["wrong_choice"])


def get_callback(activity_id: str):
    """ Given an activity id, this returns the callback corresponding to that activity. """
    return _my_callbacks[activity_id]


_my_callbacks = {
    "start": start,
    "name_nickname": name_nickname,
    "insert_name": insert_name,
    "insert_nickname": insert_nickname,
    "insert_age": insert_age,
}



from mmcc_framework.activity import ActivityType
from typing import Any, Callable, Dict, List, Union

from mmcc_framework.framework import Process, Response
from mmcc_framework.exceptions import DescriptionException


def validate_process(process:  Union[Process, Dict[str, Any]]) -> List[str]:
    """ Performs some checks on the description, both syntactic and semantic.

    If the check is successful this returns an empty list, otherwise the list
    will contain the error messages (one or more than one).

    This checks that:
    - If process is a dictionary
        - it contains the list of activities and the id of the first one
        - each activity has an id and a valid type
        - each activity provides some choices according to its type
        - every id used has a corresponding activity
    - no activity is linked to itself as the next or as a choice
    - no activity contains None or Null in the choices
    - there are not duplicate choices
    - there are no activities with the same id
    - the first activity id has a corresponding activity

    This does not check:
    - the knowledge base, its contents and how it is used
    - the callbacks, their existence and behavior
    """
    if not isinstance(process, Process):
        process = Process.from_dict(process)

    errors = []
    try:
        process.check()
    except (DescriptionException, KeyError) as err:
        errors.append(str(err))
    return errors


def validate_callbacks(
        process:  Union[Process, Dict[str, Any]],
        callback_getter: Callable[
            [str],
            Callable[[Dict[str, Any], Dict[str, Any], Dict[str, Any]], Response]
        ]):
    """ Checks that each callback used in the process is available.

    The process must be valid. ActivityType.END activities do not have a callback.

    This does not check:
    - the knowledge base, its contents and how it is used
    - that the callbacks have the correct method signature nor their behavior

    If the check is successful this returns an empty list, otherwise the list
    will contain the error messages (one or more than one).
    """
    if not isinstance(process, Process):
        process = Process.from_dict(process)

    errors = []
    for a in process.activities:
        if not a.type == ActivityType.END:
            try:
                callback = callback_getter(a.callback)
                if not callable(callback):
                    errors.append(f"The callback getter for {a.callback} in activity {a.id} "
                                  f"returned something that is not callable: {callback}")
            except BaseException as err:
                errors.append(f"The callback getter for {a.callback} in activity {a.id} "
                              f"raised an error: {str(err)}")
    return errors



from unittest.case import TestCase

from mmcc_framework.exceptions import CallbackException, DescriptionException


class TestDescriptionException(TestCase):
    def setUp(self) -> None:
        self.cause = "cause"
        self.message = "message"
        self.value = DescriptionException(self.cause, self.message)

    def test_init(self):
        self.assertIs(self.value.cause, self.cause, "The cause is correctly set")
        self.assertIs(self.value.args[0], self.message, "The message is correctly set")

    def test_str(self):
        self.assertEqual(self.value.__str__(), f"{self.message} The cause of the exception was: {self.cause}")


class TestCallbackException(TestCase):
    def setUp(self) -> None:
        self.cause = "cause"
        self.message = "message"
        self.value = CallbackException(self.cause, self.message)

    def test_init(self):
        self.assertIs(self.value.cause, self.cause, "The cause is correctly set")
        self.assertIs(self.value.args[0], self.message, "The message is correctly set")

    def test_str(self):
        self.assertEqual(self.value.__str__(), f"{self.message} The parameter of the function was: {self.cause}")



from unittest.case import TestCase

from mmcc_framework.exceptions import DescriptionException
from mmcc_framework.process import Process
from mmcc_framework.activity import Activity, ActivityType


class TestProcess(TestCase):
    def test_init(self):
        my_activities = [Activity("one", "two", ActivityType.TASK),
                         Activity("two", None, ActivityType.OR, choices=["one"])]
        my_process = Process(my_activities, "one")
        self.assertEqual(my_process.activities, my_activities, "The activities are saved")
        self.assertEqual(my_process.first, my_activities[0], "The first activity is saved")

    def test_init_with_dict(self):
        my_activities = [Activity("one", "two", ActivityType.TASK),
                         Activity("two", None, ActivityType.OR, choices=["one"])]
        my_dicts = [
            {"my_id": my_activities[0].id, "next_id": my_activities[0].next_id, "my_type": my_activities[0].type.value},
            {"my_id": my_activities[1].id, "next_id": my_activities[1].next_id, "my_type": my_activities[1].type.value, "choices": my_activities[1].choices}]
        my_process = Process(my_dicts, "one")
        self.assertEqual(my_process.activities, my_activities, "The activities are saved")
        self.assertEqual(my_process.first, my_activities[0], "The first activity is saved")

    def test_init_with_missing_first(self):
        with self.assertRaises(DescriptionException, msg="Raise if first activity id has no corresponding activity"):
            Process([Activity("one", None, ActivityType.TASK)], "two").check()

    def test_from_dict(self):
        my_activities = [Activity("one", "two", ActivityType.TASK),
                         Activity("two", None, ActivityType.OR, choices=["one"])]
        my_dict = {
            "activities": [
                {"my_id": my_activities[0].id, "next_id": my_activities[0].next_id, "my_type": my_activities[0].type.value},
                {"my_id": my_activities[1].id, "next_id": my_activities[1].next_id, "my_type": my_activities[1].type.value, "choices": my_activities[1].choices}],
                "first_activity_id": my_activities[0].id
        }
        my_process = Process.from_dict(my_dict)
        self.assertEqual(my_process.activities, my_activities, "The activities are correctly set")
        self.assertEqual(my_process.first, my_activities[0], "The first activity is correctly set")

    def test_from_dict_shuffled(self):
        my_activities = [Activity("one", "two", ActivityType.TASK),
                         Activity("two", None, ActivityType.OR, choices=["one"])]
        my_dict = {
            "first_activity_id": my_activities[0].id,
            "activities": [
                {"my_id": my_activities[0].id, "next_id": my_activities[0].next_id, "my_type": my_activities[0].type.value},
                {"my_id": my_activities[1].id, "next_id": my_activities[1].next_id, "my_type": my_activities[1].type.value, "choices": my_activities[1].choices}]
        }
        my_process = Process.from_dict(my_dict)
        self.assertEqual(my_process.activities, my_activities, "The activities are correctly set")
        self.assertEqual(my_process.first, my_activities[0], "The first activity is correctly set")

    def test_from_dict_with_wrong_params(self):
        with self.assertRaises(DescriptionException, msg="If necessary params are missing, raise. activities"):
            Process.from_dict({"first_activity_id": "one"}).check()

        with self.assertRaises(DescriptionException, msg="If necessary params are missing, raise. first id"):
            Process.from_dict({"activities": [{"my_id": "one", "next_id": "two", "my_type": "task"}]}).check()

        with self.assertRaises(DescriptionException, msg="With more params than needed, raise"):
            Process.from_dict(
                {"activities": [{"my_id": "one", "next_id": "two", "my_type": "task"}], "first_activity_id": "one",
                 "other": True}).check()

    def test_check_first_with_more_correspondences(self):
        with self.assertRaises(DescriptionException, msg="Raise if first activity id has more correspondences"):
            Process([Activity("one", None, ActivityType.TASK),
                     Activity("one", None, ActivityType.TASK)], "one").check()

    def test_check_exists_unique_next_id(self):
        with self.assertRaises(DescriptionException, msg="Raise if a next id is not unique"):
            Process([Activity("one", "two", ActivityType.TASK),
                     Activity("two", None, ActivityType.TASK),
                     Activity("two", None, ActivityType.TASK)], "one").check()

        with self.assertRaises(DescriptionException, msg="Raise if a next id is not found"):
            Process([Activity("one", "two", ActivityType.TASK)], "one").check()

    def test_check_next_id_is_not_self(self):
        with self.assertRaises(DescriptionException, msg="Raise if a next id is equal to id"):
            Process([Activity("one", "one", ActivityType.TASK)], "one").check()

    def test_check_exists_unique_choices(self):
        with self.assertRaises(DescriptionException, msg="Raise if a choice is not unique"):
            Process([Activity("one", None, ActivityType.OR, ["two"]),
                     Activity("two", None, ActivityType.TASK),
                     Activity("two", None, ActivityType.TASK)], "one").check()

        with self.assertRaises(DescriptionException, msg="Raise if a choice is not found"):
            Process([Activity("one", None, ActivityType.OR, ["two"])], "one").check()

    def test_check_choice_not_allowed(self):
        with self.assertRaises(DescriptionException, msg="Raise if a choice is None"):
            Process([Activity("one", None, ActivityType.OR, ["two", None]),
                     Activity("two", None, ActivityType.TASK)], "one").check()

        with self.assertRaises(DescriptionException, msg="Raise if a choice is the same as the id"):
            Process([Activity("one", None, ActivityType.OR, ["one"])], "one").check()

        with self.assertRaises(DescriptionException, msg="Raise if a choice contains duplicates"):
            Process([Activity("one", None, ActivityType.OR, ["two", "two"]),
                     Activity("two", None, ActivityType.TASK)], "one").check()



import tkinter as tk

from examples.rasa.my_callbacks import get_callback
from mmcc_framework import Framework, RasaNlu, validate_process, validate_callbacks


# A helper function that extracts the information from the payload of the response and returns an updated state.
def get_state(response, old_state):
    response = response["payload"]
    return {
        "show_name": response["show_name"] if "show_name" in response else old_state["show_name"],
        "show_age": response["show_age"] if "show_age" in response else old_state["show_age"],
        "show_choose_name": response["show_choose_name"] if "show_choose_name" in response else old_state[
            "show_choose_name"],
        "show_choose_nickname": response["show_choose_nickname"] if "show_choose_nickname" in response else old_state[
            "show_choose_nickname"],
        "show_field": response["show_field"] if "show_field" in response else old_state["show_field"],
        "field_contents": response["field_contents"] if "field_contents" in response else old_state["field_contents"],
    }


# A frame with some buttons and a chat.
class Application(tk.Frame):
    def __init__(self, master=None):
        super().__init__(master)

        # Prepare the state and the framework.
        self.state = {"show_name": False, "show_age": False, "show_choose_name": True, "show_choose_nickname": True,
                      "show_field": False, "field_contents": ""}
        self.framework = Framework.from_file("my_process.json",
                                             "my_kb.json",
                                             {},
                                             get_callback,
                                             RasaNlu())
        self.master = master
        self.pack()

        # Check that the process and the callbacks are valid.
        errors = validate_process(self.framework._process) + validate_callbacks(self.framework._process, get_callback)
        if errors:
            print(errors)
            exit()

        # Add the buttons.
        self.name_button = tk.Button(self.master, text='Insert "Bob"', command=lambda: self.send_data(
            RasaNlu.dict("choose_name_nickname", {"name": "Bob"})))
        self.name_button.pack()
        self.age_button = tk.Button(self.master, text='Insert "50"',
                                    command=lambda: self.send_data(RasaNlu.dict("choose_age", {"age": "50"})))
        self.age_button.pack()
        self.choose_name_button = tk.Button(self.master, text="Use the name",
                                            command=lambda: self.send_data(RasaNlu.dict("choose_name")))
        self.choose_name_button.pack()
        self.choose_nickname_button = tk.Button(self.master, text="Use the nickname",
                                                command=lambda: self.send_data(RasaNlu.dict("choose_nickname")))
        self.choose_nickname_button.pack()
        self.field = tk.Label(self.master, text="")
        self.field.pack()

        # Add the chat.
        self.chat = tk.Text(self.master, state=tk.constants.DISABLED)
        self.chat.tag_config('user_input', foreground="blue")
        self.chat.pack()
        self.insert = tk.Text(self.master, wrap='none', height=1)
        self.insert.pack()

        # The user can send the text input either pressing the button or pressing Return.
        self.go = tk.Button(self.master, text="ENTER", command=self.send_text)
        self.master.bind('<Return>', self.send_text)
        self.go.pack()

        # Update the view
        self.update_view(self.framework.handle_text_input(""))

    # Updates the view using the state.
    def update_view(self, response):
        self.chat["state"] = tk.constants.NORMAL
        self.chat.insert(tk.constants.END, "\n" + response["utterance"])
        self.chat["state"] = tk.constants.DISABLED
        self.state = get_state(response, self.state)

        self.name_button['state'] = (tk.constants.NORMAL if self.state["show_name"] else tk.constants.DISABLED)
        self.age_button["state"] = tk.constants.NORMAL if self.state["show_age"] else tk.constants.DISABLED
        self.choose_name_button["state"] = tk.constants.NORMAL if self.state[
            "show_choose_name"] else tk.constants.DISABLED
        self.choose_nickname_button["state"] = tk.constants.NORMAL if self.state[
            "show_choose_nickname"] else tk.constants.DISABLED
        self.field["state"] = tk.constants.NORMAL if self.state["show_field"] else tk.constants.DISABLED
        self.field["text"] = self.state["field_contents"]

    def send_data(self, data):
        self.update_view(self.framework.handle_data_input(data))

    def send_text(self, event=None):
        text = self.insert.get('1.0', 'end-1c')
        self.insert.delete('1.0', tk.constants.END)
        self.chat["state"] = tk.constants.NORMAL
        self.chat.insert(tk.constants.END, "\n-- " + text, "user_input")
        self.chat["state"] = tk.constants.DISABLED
        response = self.framework.handle_text_input(text)
        self.update_view(response)


# Run the app.
if __name__ == '__main__':
    root = tk.Tk()
    app = Application(master=root)
    app.mainloop()



from unittest import TestCase

from mmcc_framework.activity import Activity, ActivityType
from mmcc_framework.framework import Framework, Response, CTX_COMPLETED
from mmcc_framework.process import Process
from mmcc_framework.nlu_adapters import NoNluAdapter


class TestFrameworkBehaviourOR(TestCase):
    @staticmethod
    def callback_getter(_):
        return lambda d, k, c: Response(k, c, True, choice=d["choice"] if "choice" in d else "")

    def setUp(self) -> None:
        self.my_framework = Framework(
            Process(first_activity_id="start",
                    activities=[Activity("start", "gateway", ActivityType.START),
                                Activity("gateway", "end", ActivityType.OR, ["xor", "par", "or"]),
                                Activity("xor", None, ActivityType.XOR, ["A", "B"]),
                                Activity("A", None, ActivityType.TASK),
                                Activity("B", None, ActivityType.TASK),
                                Activity("par", None, ActivityType.PARALLEL, ["C", "D"]),
                                Activity("C", None, ActivityType.TASK),
                                Activity("D", None, ActivityType.TASK),
                                Activity("or", None, ActivityType.OR, ["E", "F"]),
                                Activity("E", None, ActivityType.TASK),
                                Activity("F", None, ActivityType.END),
                                Activity("end", None, ActivityType.END)]),
            {"my_key": "a value"},
            {"ctx_key": "ctx value"},
            self.callback_getter,
            NoNluAdapter([]),
            lambda k: None)

    def test_path_1(self):
        path = [["", "gateway", []],
                ["xor", "xor", ["gateway"]],
                ["A", "A", ["gateway"]],
                ["", "gateway", ["gateway"]],
                ["par", "par", ["gateway"]],
                ["C", "C", ["gateway"]],
                ["", "par", ["gateway"]],
                ["D", "D", ["gateway", "par"]],
                ["", "par", ["gateway", "par"]],
                [None, "gateway", ["gateway"]],
                ["or", "or", ["gateway"]],
                ["E", "E", ["gateway", "or"]],
                ["", "or", ["gateway", "or"]],
                [None, "gateway", ["gateway"]],
                ["or", "or", ["gateway"]],
                ["F", "F", ["gateway", "or"]],
                ["", "F", ["gateway", "or"]]]
        for index, step in enumerate(path):
            self.my_framework.handle_data_input({"choice": step[0]} if step[0] != "" else {"data": "value"})
            self.assertEqual(self.my_framework._current.id, step[1], f"Step index when test failed: {index}, "
                                                                     f"expected: {step[1]}")
            self.assertEqual(self.my_framework._ctx[CTX_COMPLETED], step[2], f"Step index when test failed: {index}")

    def test_path_2(self):
        path = [["", "gateway", []],
                ["xor", "xor", ["gateway"]],
                ["B", "B", ["gateway"]],
                ["", "gateway", ["gateway"]],
                ["or", "or", ["gateway"]],
                ["E", "E", ["gateway", "or"]],
                ["", "or", ["gateway", "or"]],
                [None, "gateway", ["gateway"]],
                ["par", "par", ["gateway"]],
                ["D", "D", ["gateway"]],
                ["", "par", ["gateway"]],
                ["C", "C", ["gateway", "par"]],
                ["", "par", ["gateway", "par"]],
                ["D", "D", ["gateway", "par"]],
                ["", "par", ["gateway", "par"]],
                [None, "gateway", ["gateway"]],
                [None, "end", []],
                ["", "end", []]]
        for index, step in enumerate(path):
            self.my_framework.handle_data_input({"choice": step[0]} if step[0] != "" else {"data": "value"})
            self.assertEqual(self.my_framework._current.id, step[1], f"Step index when test failed: {index}, "
                                                                     f"expected: {step[1]}")
            self.assertEqual(self.my_framework._ctx[CTX_COMPLETED], step[2], f"Step index when test failed: {index}")


class TestFrameworkBehaviourPAR(TestCase):
    @staticmethod
    def callback_getter(_):
        return lambda d, k, c: Response(k, c, True, choice=d["choice"] if "choice" in d else "")

    def setUp(self) -> None:
        self.my_framework = Framework(
            Process(first_activity_id="start",
                    activities=[Activity("start", "gateway", ActivityType.START),
                                Activity("gateway", "end", ActivityType.PARALLEL, ["xor", "par", "or"]),
                                Activity("xor", None, ActivityType.XOR, ["A", "B"]),
                                Activity("A", None, ActivityType.TASK),
                                Activity("B", None, ActivityType.TASK),
                                Activity("par", None, ActivityType.PARALLEL, ["C", "D"]),
                                Activity("C", None, ActivityType.TASK),
                                Activity("D", None, ActivityType.TASK),
                                Activity("or", None, ActivityType.OR, ["E", "F"]),
                                Activity("E", None, ActivityType.TASK),
                                Activity("F", None, ActivityType.END),
                                Activity("end", None, ActivityType.END)]),
            {"my_key": "a value"},
            {"ctx_key": "ctx value"},
            self.callback_getter,
            NoNluAdapter([]),
            lambda k: None)

    def test_path_1(self):
        path = [["", "gateway", []],
                ["xor", "xor", []],
                ["A", "A", []],
                ["", "gateway", []],
                ["par", "par", []],
                ["C", "C", []],
                ["", "par", []],
                ["D", "D", ["par"]],
                ["", "par", ["par"]],
                [None, "gateway", []],
                ["or", "or", ["gateway"]],
                ["E", "E", ["gateway", "or"]],
                ["", "or", ["gateway", "or"]],
                [None, "gateway", ["gateway"]],
                ["or", "or", ["gateway"]],
                ["F", "F", ["gateway", "or"]],
                ["", "F", ["gateway", "or"]]]
        for index, step in enumerate(path):
            self.my_framework.handle_data_input({"choice": step[0]} if step[0] != "" else {"data": "value"})
            self.assertEqual(self.my_framework._current.id, step[1], f"Step index when test failed: {index}, "
                                                                     f"expected: {step[1]}")
            self.assertEqual(self.my_framework._ctx[CTX_COMPLETED], step[2], f"Step index when test failed: {index}")

    def test_path_2(self):
        path = [["", "gateway", []],
                ["xor", "xor", []],
                ["B", "B", []],
                ["", "gateway", []],
                ["or", "or", []],
                ["E", "E", ["or"]],
                ["", "or", ["or"]],
                [None, "gateway", []],
                ["par", "par", ["gateway"]],
                ["D", "D", ["gateway"]],
                ["", "par", ["gateway"]],
                ["C", "C", ["gateway", "par"]],
                ["", "par", ["gateway", "par"]],
                ["D", "D", ["gateway", "par"]],
                ["", "par", ["gateway", "par"]],
                [None, "gateway", ["gateway"]],
                [None, "end", []],
                ["", "end", []]]
        for index, step in enumerate(path):
            self.my_framework.handle_data_input({"choice": step[0]} if step[0] != "" else {"data": "value"})
            self.assertEqual(self.my_framework._current.id, step[1], f"Step index when test failed: {index}, "
                                                                     f"expected: {step[1]}")
            self.assertEqual(self.my_framework._ctx[CTX_COMPLETED], step[2], f"Step index when test failed: {index}")


class TestFrameworkBehaviourXOR(TestCase):
    @staticmethod
    def callback_getter(_):
        return lambda d, k, c: Response(k, c, True, choice=d["choice"] if "choice" in d else "")

    def setUp(self) -> None:
        self.my_framework = Framework(
            Process(first_activity_id="start",
                    activities=[Activity("start", "gateway", ActivityType.START),
                                Activity("gateway", "end", ActivityType.XOR, ["xor", "par", "or"]),
                                Activity("xor", None, ActivityType.XOR, ["A", "B"]),
                                Activity("A", None, ActivityType.TASK),
                                Activity("B", None, ActivityType.TASK),
                                Activity("par", None, ActivityType.PARALLEL, ["C", "D"]),
                                Activity("C", None, ActivityType.TASK),
                                Activity("D", None, ActivityType.TASK),
                                Activity("or", None, ActivityType.OR, ["E", "F"]),
                                Activity("E", None, ActivityType.TASK),
                                Activity("F", None, ActivityType.END),
                                Activity("end", None, ActivityType.END)]),
            {"my_key": "a value"},
            {"ctx_key": "ctx value"},
            self.callback_getter,
            NoNluAdapter([]),
            lambda k: None)

    def test_path_1a(self):
        path = [["", "gateway", []],
                ["xor", "xor", []],
                ["A", "A", []],
                ["", "end", []],
                ["", "end", []]]
        for index, step in enumerate(path):
            self.my_framework.handle_data_input({"choice": step[0]} if step[0] != "" else {"data": "value"})
            self.assertEqual(self.my_framework._current.id, step[1], f"Step index when test failed: {index}, "
                                                                     f"expected: {step[1]}")
            self.assertEqual(self.my_framework._ctx[CTX_COMPLETED], step[2], f"Step index when test failed: {index}")

    def test_path_1b(self):
        path = [["", "gateway", []],
                ["par", "par", []],
                ["C", "C", []],
                ["", "par", []],
                ["D", "D", ["par"]],
                ["", "par", ["par"]],
                [None, "end", []],
                ["", "end", []]]
        for index, step in enumerate(path):
            self.my_framework.handle_data_input({"choice": step[0]} if step[0] != "" else {"data": "value"})
            self.assertEqual(self.my_framework._current.id, step[1], f"Step index when test failed: {index}, "
                                                                     f"expected: {step[1]}")
            self.assertEqual(self.my_framework._ctx[CTX_COMPLETED], step[2], f"Step index when test failed: {index}")

    def test_path_1c(self):
        path = [["", "gateway", []],
                ["or", "or", []],
                ["F", "F", ["or"]],
                ["", "F", ["or"]]]
        for index, step in enumerate(path):
            self.my_framework.handle_data_input({"choice": step[0]} if step[0] != "" else {"data": "value"})
            self.assertEqual(self.my_framework._current.id, step[1], f"Step index when test failed: {index}, "
                                                                     f"expected: {step[1]}")
            self.assertEqual(self.my_framework._ctx[CTX_COMPLETED], step[2], f"Step index when test failed: {index}")

    def test_path_2a(self):
        path = [["", "gateway", []],
                ["par", "par", []],
                ["D", "D", []],
                ["", "par", []],
                ["C", "C", ["par"]],
                ["", "par", ["par"]],
                ["D", "D", ["par"]],
                ["", "par", ["par"]],
                [None, "end", []],
                ["", "end", []]]
        for index, step in enumerate(path):
            self.my_framework.handle_data_input({"choice": step[0]} if step[0] != "" else {"data": "value"})
            self.assertEqual(self.my_framework._current.id, step[1], f"Step index when test failed: {index}, "
                                                                     f"expected: {step[1]}")
            self.assertEqual(self.my_framework._ctx[CTX_COMPLETED], step[2], f"Step index when test failed: {index}")


class TestFramework(TestCase):
    @staticmethod
    def callback_getter(_):
        return lambda d, k, c: Response(k, c, True)

    def setUp(self) -> None:
        self.my_kb = {"my_key": "a value"}
        self.my_ctx = {"ctx_key": "ctx value"}
        self.my_callback = self.callback_getter
        self.my_nlu = NoNluAdapter([])
        self.my_save = lambda k: None

    def test_init(self):
        my_proc = Process([Activity("one", None, ActivityType.START)], "one")

        my_framework = Framework(my_proc, self.my_kb, self.my_ctx, self.my_callback, self.my_nlu, self.my_save)
        self.assertEqual(my_framework._process, my_proc)
        self.assertEqual(my_framework._kb, self.my_kb)
        self.assertEqual(my_framework._ctx, self.my_ctx)
        self.assertEqual(my_framework._callback_getter, self.my_callback)
        self.assertEqual(my_framework._nlu, self.my_nlu)
        self.assertEqual(my_framework._on_save, self.my_save)

    def test_init_dict(self):
        proc_dic = {
            "first_activity_id": "one",
            "activities": [{"my_id": "one", "next_id": None, "my_type": "task"}]
        }
        my_framework = Framework(proc_dic, self.my_kb, self.my_ctx, self.my_callback, self.my_nlu, self.my_save)
        self.assertEqual(my_framework._process.first.id, proc_dic["first_activity_id"])
        self.assertEqual(my_framework._process.activities[0].id, proc_dic["activities"][0]["my_id"])
        self.assertEqual(my_framework._process.activities[0].next_id, proc_dic["activities"][0]["next_id"])
        self.assertEqual(my_framework._process.activities[0].type.value, proc_dic["activities"][0]["my_type"])
        self.assertEqual(my_framework._kb, self.my_kb)
        self.assertEqual(my_framework._ctx, self.my_ctx)
        self.assertEqual(my_framework._callback_getter, self.my_callback)
        self.assertEqual(my_framework._nlu, self.my_nlu)
        self.assertEqual(my_framework._on_save, self.my_save)

    def test_init_callback(self):
        my_proc = Process([Activity("one", None, ActivityType.START)], "one")

        def proc_call():
            return my_proc

        def kb_call():
            return self.my_kb

        my_framework = Framework(proc_call, kb_call, self.my_ctx, self.my_callback, self.my_nlu, self.my_save)

        self.assertEqual(my_framework._process, my_proc)
        self.assertEqual(my_framework._kb, self.my_kb)
        self.assertEqual(my_framework._ctx, self.my_ctx)
        self.assertEqual(my_framework._callback_getter, self.my_callback)
        self.assertEqual(my_framework._nlu, self.my_nlu)
        self.assertEqual(my_framework._on_save, self.my_save)

    def test_init_callback_dict(self):
        proc_dic = {
            "first_activity_id": "one",
            "activities": [{"my_id": "one", "next_id": None, "my_type": "task"}]
        }

        def proc_call():
            return proc_dic

        def kb_call():
            return self.my_kb

        my_framework = Framework(proc_call, kb_call, self.my_ctx, self.my_callback, self.my_nlu, self.my_save)
        self.assertEqual(my_framework._process.first.id, proc_dic["first_activity_id"])
        self.assertEqual(my_framework._process.activities[0].id, proc_dic["activities"][0]["my_id"])
        self.assertEqual(my_framework._process.activities[0].next_id, proc_dic["activities"][0]["next_id"])
        self.assertEqual(my_framework._process.activities[0].type.value, proc_dic["activities"][0]["my_type"])
        self.assertEqual(my_framework._kb, self.my_kb)
        self.assertEqual(my_framework._ctx, self.my_ctx)
        self.assertEqual(my_framework._callback_getter, self.my_callback)
        self.assertEqual(my_framework._nlu, self.my_nlu)
        self.assertEqual(my_framework._on_save, self.my_save)


class TestResponse(TestCase):
    def setUp(self) -> None:
        self.my_kb = {"key": "value", "key2": "value2"}
        self.my_ctx = {"c key": "c value"}
        self.my_bool = True
        self.my_utt = "An utt"
        self.my_payload = {"payload": "payload value"}
        self.my_choice = "A choice"

    def test_init(self):
        my_response = Response(self.my_kb, self.my_ctx, self.my_bool, self.my_utt, self.my_payload, self.my_choice)

        self.assertEqual(my_response.kb, self.my_kb, "The kb is correctly saved")
        self.assertEqual(my_response.ctx, self.my_ctx, "The ctx is correctly saved")
        self.assertEqual(my_response.complete, self.my_bool, "Completed is correctly saved")
        self.assertEqual(my_response.utterance, self.my_utt, "The utterance is correctly saved")
        self.assertEqual(my_response.payload, self.my_payload, "The payload is correctly saved")
        self.assertEqual(my_response.choice, self.my_choice, "The choice is correctly saved")

    def test_init_with_default(self):
        my_response = Response(self.my_kb, self.my_ctx, self.my_bool)
        self.assertEqual(my_response.kb, self.my_kb, "The kb is correctly saved")
        self.assertEqual(my_response.ctx, self.my_ctx, "The ctx is correctly saved")
        self.assertEqual(my_response.complete, self.my_bool, "Completed is correctly saved")
        self.assertEqual(my_response.utterance, "", "The default utterance is empty")
        self.assertEqual(my_response.payload, {}, "The default payload is empty")
        self.assertIsNone(my_response.choice, "The default choice is None")

    def test_to_dict(self):
        my_dict = Response(self.my_kb, self.my_ctx, self.my_bool, self.my_utt, self.my_payload,
                           self.my_choice).to_dict()

        self.assertEqual(my_dict["utterance"], self.my_utt, "The utterance is correctly passed to the dict")
        self.assertEqual(my_dict["payload"], self.my_payload, "The payload is correctly passed to the dict")

    def test_to_dict_with_default(self):
        my_dict = Response(self.my_kb, self.my_ctx, self.my_bool).to_dict()

        self.assertEqual(my_dict["utterance"], "", "The utterance is correctly passed to the dict")
        self.assertEqual(my_dict["payload"], {}, "The payload is correctly passed to the dict")

    def test_add_utterance(self):
        my_response = Response(self.my_kb, self.my_ctx, self.my_bool, self.my_utt, self.my_payload, self.my_choice)
        my_response.add_utterance(self.my_kb, "key")
        self.assertEqual(my_response.utterance, self.my_utt + "\n" + self.my_kb["key"])

        my_response = Response(self.my_kb, self.my_ctx, self.my_bool, self.my_utt, self.my_payload, self.my_choice)
        my_response.add_utterance(self.my_kb, "key3", "def")
        self.assertEqual(my_response.utterance, self.my_utt + "\n" + "def")

        my_response = Response(self.my_kb, self.my_ctx, self.my_bool, self.my_utt, self.my_payload, self.my_choice)
        my_response.add_utterance(self.my_kb, "key3")
        self.assertEqual(my_response.utterance, self.my_utt)

        my_response = Response(self.my_kb, self.my_ctx, self.my_bool)
        my_response.add_utterance(self.my_kb, "key")
        self.assertEqual(my_response.utterance, self.my_kb["key"])



class DescriptionException(Exception):
    """ Exception raised when the check on the process description finds errors or incongruities.

    :ivar cause: the element that caused the error
    :ivar message: the message of this exception
    """

    def __init__(self, cause, message: str = "The process description caused an exception.") -> None:
        """ Creates an exception with the provided cause, and an optional message. """
        super().__init__(message)
        self.cause = cause

    def __str__(self) -> str:
        """ Presents the message and the cause of this exception. """
        return f"{super().__str__()} The cause of the exception was: {self.cause}"


class CallbackException(Exception):
    """ Exception raised when the check on the callbacks fails.

    :ivar cause: the parameter that caused the exception
    :ivar message: the message of this exception
    """

    def __init__(self, cause, message: str = "A callback caused an exception.") -> None:
        """ Creates an exception with the provided cause, and an optional message. """
        super().__init__(message)
        self.cause = cause

    def __str__(self) -> str:
        """ Presents the message and the cause of this exception. """
        return f"{super().__str__()} The parameter of the function was: {self.cause}"



from mmcc_framework.activity import Activity, ActivityType
from mmcc_framework.exceptions import CallbackException, DescriptionException
from mmcc_framework.framework import CTX_COMPLETED, Framework, Response
from mmcc_framework.nlu_adapters import NoNluAdapter, RasaNlu
from mmcc_framework.process import Process
from mmcc_framework.validation import validate_process, validate_callbacks



from enum import Enum
from typing import Any, Dict, List, Optional, Union

from mmcc_framework.exceptions import DescriptionException


class Activity(object):
    """ An element of a `Process` description, a single step in which the user has to do something.

    :ivar id: the unique id of this activity (unique)
    :ivar next_id: the id of the `Activity` that comes after this, when completed (can be `None`)
    :ivar type: the `ActivityType` of this activity
    :ivar choices: a list of id that this activity offers as choices (can be `None`)
    :ivar callback: the callback identifier for this activity
    """

    def __init__(self,
                 my_id: str,
                 next_id: Optional[str],
                 my_type: Union[str, "ActivityType"],
                 choices: List[str] = None,
                 callback: Optional[str] = None) -> None:
        """ Creates a new activity with the provided parameters; performs some checks.

        The parameter `next_id` is `None` if the activity is the last "inside" one of the
        `ActivityType.get_require_choice()` gateways.

        The `choices` are required only if `my_type` is in `ActivityType.get_require_choice()`.

        The action used by this activity is provided as a separate callback. This callback is
        identified with the `callback` attribute; if the attribute is not provided, `my_id` will
        be used instead.

        :param my_id: the unique id of this activity
        :param next_id: the id of the `Activity` that comes after this, when completed
        :param my_type: the `ActivityType` of this activity or a string representing it
        :param choices: a list of id that this activity offers as choices
        :param callback: the callback identifier if different from `id`
        :raises DescriptionException: if `choices` is provided and not needed, or is missing
        :raises KeyError: if can not recognize the `ActivityType` provided
        """
        self.id = my_id
        self.next_id = next_id
        self.type = my_type if isinstance(
            my_type, ActivityType) else ActivityType[my_type.upper()]
        if self.type in ActivityType.get_require_choice():
            if choices is None:
                raise DescriptionException(
                    self.id, "Expected some choices, but found none.")
            if not choices:
                raise DescriptionException(
                    self.id, "Expected some choices, but found an empty list.")
        elif choices is not None or choices:
            raise DescriptionException(self.id, "Found unexpected choices.")
        self.choices = choices
        self.callback = my_id if callback is None else callback

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Any]) -> "Activity":
        """ Returns the `Activity` represented by a given dictionary, if possible.

        Example:
            my_activity = Activity.from_dict({"my_id": "an id",
                                              "next_id": "another id",
                                              "my_type": "xor",
                                              "choices": ["id", "more id"],
                                              "callback": "second_callback"})

        See `Activity.__init__` for more info.

        :param dictionary: a dictionary that represents an `Activity`
        :return: an `Activity` instance with the provided attributes
        :raises DescriptionException: if parameters are missing, or unknown parameters are provided
        """
        try:
            return cls(**dictionary)
        except TypeError as err:
            raise DescriptionException(
                dictionary, "Did not find a required parameter in this activity.") from err

    def __eq__(self, o: object) -> bool:
        """ Returns `True` if two activities have the same attributes. """
        return isinstance(o, Activity) and \
            self.id == o.id and self.next_id == o.next_id and self.type == o.type \
            and self.choices == o.choices and self.callback == o.callback

    def __ne__(self, o: object) -> bool:
        """ Returns `True` if `__eq__` would return `False`. """
        return not self == o


class ActivityType(Enum):
    """ The various types of activities in a process. """

    TASK = "task"
    """ Represents an operation to be done to complete the process.
    The `Response` contains `True` if the user can move on to the next activity.
    """
    START = "start"
    """ It is the entry point of the process.
    Its `Response` must contain `True` and can be used to prepare the state using the payload.
    """
    END = "end"
    """ A "sink" state that represents the termination of the process. """
    PARALLEL = "parallel"
    """ A task that gives some options to the user, the user can choose which one to execute.
    This is completed when all options have been chosen at least once.
    Its callback must return the id of the chosen activity if the user input was valid, `None` to
    move on to the next task.
    """
    XOR = "xor"
    """ A task that gives some options to the user, the user can choose which one to execute.
    This allows the user to choose exactly one of the options.
    Its callback must return the id of the chosen activity if the user input was valid.
    """
    OR = "or"
    """ A task that gives some options to the user, the user can choose which one to execute.
    This is completed when the user has chosen at least one of the options.
    Its callback must return the id of the chosen activity if the user input was valid, `None` to
    move on to the next task.
    """

    @staticmethod
    def get_require_choice() -> List["ActivityType"]:
        """ Returns a list of the `ActivityType`s that require the choices in the description. """
        return [ActivityType.PARALLEL, ActivityType.XOR, ActivityType.OR]



from unittest.case import TestCase

from mmcc_framework.exceptions import DescriptionException
from mmcc_framework.activity import Activity, ActivityType


class TestActivity(TestCase):
    def setUp(self) -> None:
        self.my_id = "my id"
        self.next_id = "next id"
        self.choices = ["a", "b", "c"]
        self.callback = "a callback"
        self.value = {}

    def test_init(self):
        for t in ActivityType:
            self.value[t] = Activity(self.my_id, self.next_id, t, self.choices if t in ActivityType.get_require_choice() else None)

        for t in ActivityType:
            self.assertEqual(self.value[t].id, self.my_id, "The id is correctly set")
            self.assertEqual(self.value[t].next_id, self.next_id, "The next id is correctly set")
            self.assertEqual(self.value[t].type, t, "The type is correctly set")
            self.assertEqual(self.value[t].choices, self.choices if t in ActivityType.get_require_choice() else None, "The choices are correctly set")
            self.assertEqual(self.value[t].callback, self.my_id, "The callback is correctly set")

    def test_init_with_type_from_text(self):
        for t in ActivityType:
            self.value[t] = Activity(self.my_id, self.next_id, t.value, self.choices if t in ActivityType.get_require_choice() else None)

        for t in ActivityType:
            self.assertEqual(self.value[t].id, self.my_id, "The id is correctly set")
            self.assertEqual(self.value[t].next_id, self.next_id, "The next id is correctly set")
            self.assertEqual(self.value[t].type, t, "The type is correctly casted and set")
            self.assertEqual(self.value[t].choices, self.choices if t in ActivityType.get_require_choice() else None, "The choices are correctly set")
            self.assertEqual(self.value[t].callback, self.my_id, "The callback is correctly set")

    def test_init_with_callback(self):
        c_name = "a callback"
        for t in ActivityType:
            self.value[t] = Activity(self.my_id, self.next_id, t.value, self.choices if t in ActivityType.get_require_choice() else None, c_name)

        for t in ActivityType:
            self.assertEqual(self.value[t].id, self.my_id, "The id is correctly set")
            self.assertEqual(self.value[t].next_id, self.next_id, "The next id is correctly set")
            self.assertEqual(self.value[t].type, t, "The type is correctly casted and set")
            self.assertEqual(self.value[t].choices, self.choices if t in ActivityType.get_require_choice() else None, "The choices are correctly set")
            self.assertEqual(self.value[t].callback, c_name, "The callback is correctly set")

    def test_init_with_wrong_choices(self):
        for t in ActivityType:
            with self.assertRaises(DescriptionException, msg="If wrong choices are provided, should raise exception"):
                Activity(self.my_id, self.next_id, t, None if t in ActivityType.get_require_choice() else self.choices)

    def test_init_with_wrong_type(self):
        with self.assertRaises(KeyError, msg="If the type is wrong, should raise an exception"):
            Activity(self.my_id, self.next_id, "no_type")

    def test_from_dict(self):
        for t in ActivityType:
            my_dict = {"my_id": self.my_id, "next_id": self.next_id, "my_type": t.value}
            if t in ActivityType.get_require_choice():
                my_dict["choices"] = self.choices
            self.value[t] = Activity.from_dict(my_dict)

        for t in ActivityType:
            self.assertEqual(self.value[t].id, self.my_id, "The id is correctly set")
            self.assertEqual(self.value[t].next_id, self.next_id, "The next id is correctly set")
            self.assertEqual(self.value[t].type, t, "The type is correctly set")
            self.assertEqual(self.value[t].choices, self.choices if t in ActivityType.get_require_choice() else None, "The choices are correctly set")
            self.assertEqual(self.value[t].callback, self.my_id, "The callback is correctly set")

    def test_from_dict_shuffled(self):
        for t in ActivityType:
            my_dict = {"next_id": self.next_id, "my_id": self.my_id, }
            my_dict["callback"] = self.callback
            if t in ActivityType.get_require_choice():
                my_dict["choices"] = self.choices
            my_dict["my_type"] = t.value
            self.value[t] = Activity.from_dict(my_dict)

        for t in ActivityType:
            self.assertEqual(self.value[t].id, self.my_id, "The id is correctly set")
            self.assertEqual(self.value[t].next_id, self.next_id, "The next id is correctly set")
            self.assertEqual(self.value[t].type, t, "The type is correctly set")
            self.assertEqual(self.value[t].choices, self.choices if t in ActivityType.get_require_choice() else None, "The choices are correctly set")
            self.assertEqual(self.value[t].callback, self.callback, "The callback is correctly set")

    def test_from_dict_with_wrong_params(self):
        with self.assertRaises(DescriptionException, msg="If necessary params are missing, raise. my_id"):
            Activity.from_dict({"next_id": self.next_id, "my_type": ActivityType.XOR, "choices": self.choices})

        with self.assertRaises(DescriptionException, msg="If necessary params are missing, raise. next_id"):
            Activity.from_dict({"my_id": self.my_id, "my_type": ActivityType.XOR, "choices": self.choices})

        with self.assertRaises(DescriptionException, msg="If necessary params are missing, raise. my_type"):
            Activity.from_dict({"my_id": self.my_id, "next_id": self.next_id, "choices": self.choices})

        with self.assertRaises(DescriptionException, msg="With more params than needed, raise"):
            Activity.from_dict({"my_id": self.my_id, "next_id": self.next_id, "my_type": ActivityType.XOR, "choices": self.choices, "other": True})

    def test_eq_ne(self):
        subject = Activity(self.my_id, self.next_id, ActivityType.TASK)
        equal = Activity(self.my_id, self.next_id, ActivityType.TASK)
        diff_id = Activity("different", self.next_id, ActivityType.TASK)
        diff_next = Activity(self.my_id, "different", ActivityType.TASK)
        diff_type = Activity(self.my_id, self.next_id, ActivityType.OR, self.choices)
        diff_callback = Activity(self.my_id, self.next_id, ActivityType.TASK, callback="callback")
        subject_choices = Activity(self.my_id, self.next_id, ActivityType.OR, ["a"])

        self.assertEqual(subject, subject, "Activity should be equal to itself")
        self.assertEqual(subject, equal, "Activities with same attributes should be equal")
        self.assertNotEqual(subject, diff_id, "Activities with different id should not be equal")
        self.assertNotEqual(subject, diff_next, "Activities with different next id should not be equal")
        self.assertNotEqual(subject, diff_type, "Activities with different type should not be equal")
        self.assertNotEqual(subject_choices, diff_type, "Activities with different choices should not be equal")
        self.assertNotEqual(subject, diff_callback, "Activities with different callbacks should not be equal")



import json
from abc import ABC, abstractmethod
from http.client import HTTPConnection
from typing import Dict, Any, List


class NluAdapter(ABC):
    """ A class that provides a method to transform the text input into the equivalent data input."""

    @abstractmethod
    def parse(self, utterance: str) -> Dict[str, Any]:
        """ Transforms the provided text input into the equivalent data input. """
        raise NotImplementedError()


class NoNluAdapter(NluAdapter):
    """ This adapter does not use a NLU engine, and simply takes the input and puts it into a dictionary.

    The list of keys to use in the dictionary must be provided to the NoNluAdapter constructor.

    Example:
        Suppose that the framework callbacks use only the following keys: "name" and "occupation".
        Initialize the adapter: `my_adapter = NoNluAdapter(["name", "occupation"])`.

        Suppose that it is time to insert the name. If it is necessary to insert it as text use:
        `my_framework.handle_text_input("Mark")`. The callback corresponding to the current activity will receive:
        `{"name": "Mark", "occupation": "Mark"}`.

        If it is necessary to insert the name as data use: `my_framework.handle_data_input({"name": "Mark"})`.

    :ivar keys: the list of keys that are used in the callbacks
    """

    def __init__(self, expected_keys: List[str]) -> None:
        """ Initializes this adapter with the provided list of keys.

        :param expected_keys: the list of keys that are used in the callbacks
        """
        self.keys = expected_keys.copy()

    def parse(self, utterance: str) -> Dict[str, Any]:
        """ Provides a dictionary containing the utterance as the value for each key in self.keys.

        :param utterance: the text input from the user
        :return: a dictionary containing the utterance as the value for each key in self.keys
        """
        return dict.fromkeys(self.keys, utterance)


class RasaNlu(NluAdapter):
    """ This adapter uses Rasa, to use this adapter it is necessary to first setup and train the interpreter.

    The instructions on how to use Rasa are available on Rasa's website, and consist basically in the following steps:

    - Install Rasa and its dependencies;
    - Run `rasa init` in your folder of choice;
    - Edit the `data/nlu` file with the utterances used for training;
    - Run `rasa train nlu` to produce a model;
    - Start rasa on port 5005 and pass the location of the model:
      for example `rasa run --enable-api -m models/nlu-20201228-183937.tar.gz`

    Example:
        Suppose that the nlu is trained with, among the others, the intent "insert_name" with a entity "name".
        Initialize the adapter: `my_adapter = RasaNlu()`

        Suppose that it is time to insert the name. If it is necessary to insert it as text use:
        `my_framework.handle_text_input("Mark")`. The callback corresponding to the current activity will receive
        (if the intent is recognized): `{"intent": "insert_name", "name": "Mark"}`.

        If it is necessary to insert the name as data use:
        `my_framework.handle_data_input(RasaNlu.dict("insert_name", {"name": "Mark"}))`, which will pass to the callback
        the same structure as above.

    :ivar interpreter: the instance of the rasa interpreter used by this adapter
    """

    def parse(self, utterance: str) -> Dict[str, Any]:
        """ Runs the interpreter to parse the given utterance and returns a dictionary containing the parsed data.

        If no intent can be extracted from the provided utterance, this returns an empty dictionary.

        :param utterance: the text input from the user
        :return: a dictionary containing the detected intent and corresponding entities if any exists.
        """
        connection = HTTPConnection("localhost:5005") # TODO modify here to use without docker
        connection.request("POST", "/model/parse", json.dumps({"text": utterance}))
        response = json.loads(connection.getresponse().read())
        if response["intent"]["name"] is None:
            return {"intent": ""}
        return self.dict(response["intent"]["name"],
                         {item['entity']: item["value"] for item in response["entities"]})

    @staticmethod
    def dict(intent: str, values: Dict[str, Any] = None) -> Dict[str, Any]:
        """ Helper method that can be used to produce a dictionary equivalent to the one of the parse method.
        Use this method with framework.handle_data_input.

        :param intent: the intent corresponding to this input
        :param values: an optional dictionary containing pairs of entity-value
        :return: a dictionary equivalent to the one produced by the parse method
        """
        if values is None:
            values = {}
        return {"intent": intent, **values}



import json
from collections import deque
from threading import Lock
from typing import Union, Dict, Any, Callable

from mmcc_framework.process import Process
from mmcc_framework.activity import ActivityType
from mmcc_framework.nlu_adapters import NluAdapter

CTX_COMPLETED = "_done_"
""" Context key whose value is a list of activity id for the pending gateways that allow skipping. """


class Framework(object):
    """ A sort of state machine, takes a process description and handles inputs, keeping track of the current activity.

    :ivar _process: an object that represents the process for this Framework instance
    :ivar _kb: the data that is saved between different process executions
    :ivar _ctx: the data that is not saved between different process executions
    :ivar _current: the activity of the process that is being executed
    :ivar _callback_getter: a function returning the callback of an activity given its callback id
    :ivar _nlu: provides a translation from text to data, to handle in the same way text and data input (multimodal)
    :ivar _on_save: a function called when it is time to save the kb
    :ivar _stack: a pile of Activity id that is used to handle the gateways
    :ivar _done: a list that is used to determine if a gateway is completed
    """

    def __init__(self,
                 process: Union["Process", Dict[str, Any], Callable[[], Union["Process", Dict[str, Any]]]],
                 kb: Union[Dict[str, Any], Callable[[], Dict[str, Any]]],
                 initial_context: Dict[str, Any],
                 callback_getter: Callable[
                     [str], Callable[[Dict[str, Any], Dict[str, Any], Dict[str, Any]], "Response"]],
                 nlu: NluAdapter,
                 on_save: Callable[[Dict[str, Any]], None]) -> None:
        """ Instantiates a Framework with the given parameters.

        The process parameter can be a Process instance or a dictionary representing a process. In alternative, it can
        also be a function that returns a Process or a dictionary.
        Similarly, the kb parameter can be a dictionary or a callback that returns a dictionary.

        :param process: the Process for this instance, a dictionary representing it, or a callable that provides it
        :param kb: the data that is saved between different process executions
        :param initial_context: can be empty or contain configuration variables
        :param callback_getter: a function that returns the callback of an activity given its id
        :param nlu: provides a translation from text to data, to handle in the same way text and data input
        :param on_save: the function called when it is time to save the kb
        """
        if callable(process):
            process = process()
        self._process = process if isinstance(process, Process) else Process.from_dict(process)
        if callable(kb):
            kb = kb()
        self._kb = kb
        self._ctx = initial_context
        self._ctx[CTX_COMPLETED] = []
        self._current = self._process.first
        self._callback_getter = callback_getter
        self._nlu = nlu
        self._on_save = on_save
        self._stack = deque()
        self._done = {}

    @classmethod
    def from_file(cls,
                  process: str,
                  kb: str,
                  initial_context: Union[str, Dict[str, Any]],
                  callback_getter: Callable[
                      [str], Callable[[Dict[str, Any], Dict[str, Any], Dict[str, Any]], "Response"]],
                  nlu: NluAdapter,
                  lock: Lock = Lock()) -> "Framework":
        """ Loads the configuration of a framework from the files provided.

        The process file must contain a Process description that will be handled by Process.fromDict().
        The kb and context files must contain a dictionary, the context can also be provided directly.
        The kb will be saved back to its file when the process is completed.
        If the possibility exists that the files will be handled by more than one Framework instance at the time, it is
        necessary to provide a unique lock shared by all the instances. This will allow the framework to correctly
        handle the concurrency.

        :param process: the path to a file containing the process description
        :param kb: the path to a file containing the kb
        :param initial_context: the context or the path to a file containing the context
        :param callback_getter: a function that returns the callback of an activity given its id
        :param nlu: provides a translation from text to data, to handle in the same way text and data input
        :param lock: a unique lock shared by all the instances that can use the files
        """
        with lock:
            if not isinstance(initial_context, dict):
                with open(initial_context) as ctx_file:
                    my_ctx = json.load(ctx_file)
            else:
                my_ctx = initial_context

            with open(process) as process_file, open(kb) as kb_file:
                my_framework = cls(json.load(process_file),
                                   json.load(kb_file),
                                   my_ctx,
                                   callback_getter,
                                   nlu,
                                   lambda kb_c: _on_file_save(kb_c, kb, lock))
        return my_framework

    def handle_text_input(self, text: str) -> Dict[str, Any]:
        """ Takes textual input from the user, uses the nlu to parse it, and handles the input as data.

        :param text: the textual input from the user, to be parsed
        :return: a dictionary containing an utterance and a payload
        """
        return self.handle_data_input(self._nlu.parse(text.rstrip()))

    def handle_data_input(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """ Takes data input from the user and handles it.

        This will call the current activity callback and pass to it the data, then will forward the returned response
        utterance and payload to the caller.
        If the callback signals that the activity was completed successfully, this moves to the next activity in the
        process.

        :param data: the data representing the input from the user, formatted accordingly to the chosen NluAdapter
        :return: a dictionary containing an utterance and a payload
        """
        # If the activity is an END, return the default utterance if it exists.
        if self._current.type == ActivityType.END:
            return Response({}, {}, True).add_utterance(self._kb, self._current.id).to_dict()

        # If the activity is a XOR, get the choice from the callback.
        if self._current.type == ActivityType.XOR:
            response = self._get_response(data)

            # If the choice is valid, push next on the stack and continue with the chosen activity.
            if response.complete:
                # Push next id on the stack, can be None.
                self._stack.append(self._current.next_id)

                # Current will not be None, because XOR must return a valid next id.
                self._current = next(x for x in self._process.activities if x.id == response.choice)

                # Add default utterance if it exists.
                response.add_utterance(self._kb, self._current.id)

                # If the task is END, save the KB.
                if self._current.type == ActivityType.END:
                    self._on_save(self._kb)
            return response.to_dict()

        # PARALLEL and OR have similar behaviour and they are handled together.
        if self._current.type == ActivityType.PARALLEL or self._current.type == ActivityType.OR:
            # Obtain the chosen task from the callback.
            response = self._get_response(data)
            if response.complete:
                # The returned task is valid, and can be None to go to the next.
                if response.choice is None:
                    # Clear the info on the current gateway, if some exist.
                    self._done.pop(self._current.id, "")
                    if self._current.id in self._ctx[CTX_COMPLETED]:
                        self._ctx[CTX_COMPLETED].remove(self._current.id)

                    # Go to next task.
                    self._go_next(response)
                else:
                    # Put the gateway on the stack.
                    self._stack.append(self._current.id)

                    # Add an entry for this gateway to done and add the choice to it.
                    if self._current.id not in self._done:
                        self._done[self._current.id] = []
                    if response.choice not in self._done[self._current.id]:
                        self._done[self._current.id].append(response.choice)

                    # Handle separately PARALLEL and OR for updating CTX_COMPLETED.
                    if self._current.type == ActivityType.PARALLEL:
                        # A PARALLEL is completed when all the sub-tasks have been chosen at least once.
                        if all(i in self._done[self._current.id] for i in self._current.choices):
                            # Completed: add to the list.
                            if self._current.id not in self._ctx[CTX_COMPLETED]:
                                self._ctx[CTX_COMPLETED].append(self._current.id)
                        else:
                            # Not completed: remove from the list.
                            if self._current.id in self._ctx[CTX_COMPLETED]:
                                self._ctx[CTX_COMPLETED].remove(self._current.id)
                    else:
                        # An OR is completed after the first valid choice.
                        if self._current.id not in self._ctx[CTX_COMPLETED]:
                            self._ctx[CTX_COMPLETED].append(self._current.id)

                    # Set the choice and optional default utterance, the choice can not be None.
                    self._current = next(x for x in self._process.activities if x.id == response.choice)
                    response.add_utterance(self._kb, self._current.id)

                    # If the task is END, save the KB.
                    if self._current.type == ActivityType.END:
                        self._on_save(self._kb)
            return response.to_dict()

        # If the activity is TASK or START, the evaluate callback is called.
        response = self._get_response(data)

        # If the activity is completed go to the next.
        if response.complete:
            self._go_next(response)
        return response.to_dict()

    def _get_response(self, data):
        # Run the callback, update the context and the kb, and return the response.
        response = self._callback_getter(self._current.callback)(data, self._kb, self._ctx)
        self._kb = response.kb
        self._ctx = response.ctx
        return response

    def _go_next(self, response):
        # Go to the next task (maybe from the stack) and add the default utterance if it exists.
        if self._current.next_id is None:
            popped = self._stack.pop()
            while popped is None:
                popped = self._stack.pop()
            self._current = next(x for x in self._process.activities if x.id == popped)
        else:
            self._current = next(x for x in self._process.activities if x.id == self._current.next_id)
        response.add_utterance(self._kb, self._current.id)

        # If the task is END, save the KB.
        if self._current.type == ActivityType.END:
            self._on_save(self._kb)

    def is_complete(self):
        """ Returns whether the current activity type is `ActivityType.END`. """
        return self._current.type == ActivityType.END


def _on_file_save(contents: Dict[str, Any], path: str, lock: Lock) -> None:
    """ The callback used to save a json formatted dictionary to a file.

    If the file is shared, provide a lock that is unique for all the instances, and this method will handle concurrent
    access to the file.

    :param contents: the dictionary to save
    :param path: the path of the destination file
    :param lock: a lock shared by all instances that have access to the file
    """
    with lock:
        with open(path, "w") as kb_file:
            json.dump(contents, kb_file, indent=2)


class Response(object):
    def __init__(self,
                 kb: Dict[str, Any],
                 ctx: Dict[str, Any],
                 complete: bool,
                 utterance: str = None,
                 payload: Dict[str, Any] = None,
                 choice: str = None) -> None:
        """ Creates a Response with the provided parameters.

        If the current activity is one of ActivityType.get_require_choice(), and is completed, the Response will contain
        the choice of the user. This must be the id of one of the choices provided in the description.

        :param kb: the updated knowledge
        :param ctx: the updated context
        :param complete: whether the current activity is completed
        :param utterance: an optional utterance to be displayed
        :param payload: an optional payload to be returned to the caller
        :param choice: if the current activity is in ActivityType.get_require_choice() this can contain the user choice
        """
        self.kb = kb
        self.ctx = ctx
        self.complete = complete
        self.utterance = utterance if utterance is not None else ""
        self.payload = payload if payload is not None else {}
        self.choice = choice

    def to_dict(self) -> Dict[str, Any]:
        """ Returns a dictionary with utterance and payload, that can be returned to the caller. """
        return {"utterance": self.utterance, "payload": self.payload}

    def add_utterance(self, kb: Dict[str, Any], key: str, fallback: str = "") -> "Response":
        """ Adds an utterance to this response.

        The utterance is taken from the kb using the provided key, if it is not present a fallback (empty by default) is
        used. If this response does not already contain an utterance, in the end it will contain the added utterance.
        If the utterance to add can not be found and a fallback is not provided, nothing is added.
        If an utterance is provided and one already exists, the new one is appended on a new line.

        :param kb: the kb from which to take the utterance to add
        :param key: the key to retrieve the utterance from the kb
        :param fallback: the value that is used if the key is not in the kb
        :return: the updated Response
        """
        my_utt = kb[key] if key in kb else fallback
        if self.utterance == "":
            self.utterance = my_utt
        elif my_utt != "":
            self.utterance += "\n" + my_utt
        return self



from typing import Any, Dict, List, Union

from mmcc_framework.exceptions import DescriptionException
from mmcc_framework.activity import Activity, ActivityType


class Process(object):
    """ The description of a process, with a list of activities and the first activity.

    :ivar activities: a list of `Activity` objects representing this process
    :ivar first: the first Activity of the process
    """

    def __init__(self, activities: List[Union["Activity", Dict[str, Any]]], first_activity_id: str) -> None:
        """ Creates a new process description with the provided activities and first activity id.

        If the provided activities list contains the activities as dictionaries instead of
        `Activity` objects, this will call `Activity.from_dict(...)` on each of them before adding
        it.

        Example:
            my_process = Process([Activity("one", "two", ActivityType.TASK), ...], "one")

        :param activities: a list of objects or dictionaries representing the process activities
        :param first_activity_id: the id of the first activity of the process
        :raises DescriptionException: if first activity id has no corresponding activity
        """
        self.activities = []
        for a in activities:
            self.activities.append(a if isinstance(
                a, Activity) else Activity.from_dict(a))
        try:
            self.first = next(
                x for x in self.activities if x.id == first_activity_id)
        except StopIteration as err:
            raise DescriptionException(
                first_activity_id, "Found no activity with the provided id.") from err

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Any]) -> "Process":
        """ Returns the `Process` represented by a given dictionary, if possible.

        Example:
            my_process = Process.from_dict({
                            "first_activity_id": "one",
                            "activities": [
                                { "my_id": "one", "next_id": "two", "my_type": "task" },
                                ...
                            ]})


        See `Process.__init__` for more info.

        :param dictionary: a dictionary that represents a `Process`
        :return: a `Process` instance with the provided attributes
        :raises DescriptionException: if parameters are missing, or unknown parameters are provided
        """
        try:
            return cls(**dictionary)
        except TypeError as err:
            raise DescriptionException(
                dictionary, "Did not find a required parameter in the process.") from err

    def check(self) -> None:
        """ Performs some checks on the description, both syntactic and semantic (for example id are unique...).

        :raises DescriptionException: if the check is not passed
        """

        # Assume the first activity id is not found.
        found_first = 0

        for a in self.activities:
            # Check that next id is not equal to id.
            if a.next_id == a.id:
                raise DescriptionException(
                    a.id, "Found an activity that is the next of itself.")

            # Count how many activities have first id as their id.
            if self.first.id == a.id:
                found_first += 1

            # If this is a OR, XOR or PARALLEL, check that choices exist unique.
            choices = {}
            if a.type in ActivityType.get_require_choice():
                # Choices list is provided because of previous checks in Activity constructor.
                for c in a.choices:
                    if c is None:
                        raise DescriptionException(
                            a.id, "Fond an activity that contains None in the choices.")
                    if c == a.id:
                        raise DescriptionException(
                            a.id, "Found an activity with itself in its choices.")
                    if c in choices:
                        raise DescriptionException(
                            a.id, "Found an activity that contains duplicate choices.")
                    choices[c] = 0

            # Also count how many other activities have this next id as their id.
            found_next = 0
            for b in self.activities:
                if b.id == a.next_id:
                    found_next += 1

                # And count how many activities correspond to a choice.
                if b.id in choices:
                    choices[b.id] += 1

            # Raise an exception if a choice does not have a corresponding activity or has more than one.
            for c, v in choices.items():
                if v == 0:
                    raise DescriptionException(
                        a.id, f"The following does not have a corresponding activity: {c}.")
                if v > 1:
                    raise DescriptionException(
                        a.id, f"The following have multiple corresponding activities: {c}.")

            # Raise exceptions if next id or first id do not have exactly one corresponding activity.
            if a.next_id is not None:
                if found_next == 0:
                    raise DescriptionException(
                        a.id, "The provided next id does not have a corresponding activity.")
                if found_next > 1:
                    raise DescriptionException(
                        a.next_id, "Found a next id that has multiple corresponding activities.")
        if found_first == 0:
            raise DescriptionException(
                self.first.id, "First activity id has no corresponding activity.")
        if found_first > 1:
            raise DescriptionException(
                self.first.id, "First activity id has multiple corresponding activities.")
